(******************************************************************************
	Layout support.
	Written by Itby. 2006-2009.
 ******************************************************************************)
unit Layouts;

interface

{.$DEFINE DECAL} // remove a point if you have DECAL

{$IFNDEF DECAL}
	//error: Should use DECAL to compile this unit.
{$ENDIF}

uses Windows, Types, DeCAL, Classes, Graphics, Controls, RectUtils;

type
	// Rectangular table with fixed cell sizes.
	TRTable = class
	private
		FColCount: Integer;
		FRowCount: Integer;
		FBoundsRect: TRect;
		function GetCell(Col, Row: Integer): TRect; virtual;
		procedure SetBoundsRect(const Value: TRect);
		procedure SetCell(Col, Row: Integer; const Value: TRect); virtual;
	public
		constructor Create(rcRect: TRect; RowCount, ColCount: Integer);
		destructor Destroy; override;

		property Cell[Col, Row : Integer]: TRect read GetCell write SetCell; default;

		property ColCount: Integer read FColCount;
		property RowCount: Integer read FRowCount;
		property BoundsRect: TRect read FBoundsRect write SetBoundsRect;
	end;

	TLTPaneNode = class;

	TLTPane = class(TComponent)
	private
		FNode: TLTPaneNode;
		procedure SetNode(const Value: TLTPaneNode);
		function GetPaintArea: TRect;
	public
		procedure Paint(Canvas: TCanvas); virtual; abstract;
		procedure Invalidate;

		constructor Create(AOwner: TComponent; ANode: TLTPaneNode = nil); reintroduce; virtual;
		destructor Destroy; override;

		property Node: TLTPaneNode read FNode write SetNode;
		property PaintArea: TRect read GetPaintArea;
	end;

	TLayoutPane = class(TLTPane)
	public
		constructor Create(AOwner: TComponent; ANode: TLTPaneNode = nil); override;
	end;

	TLTObject = class(TObject)
	private
		FBoundsRect: TRect;
	protected
		procedure SetBoundsRect(const Value: TRect); virtual;
	public
		property BoundsRect: TRect read FBoundsRect write SetBoundsRect;
		procedure Realign; virtual; abstract;
		procedure Paint(Canvas: TCanvas); virtual; abstract;
		procedure Invalidate; virtual; abstract;
	end;

	TLTNode = class(TLTObject)
	private
		FParent: TLTObject;
		FPosition: TPoint;
		FFixedSize: TSize;
		procedure SetParent(const Value: TLTObject);
		procedure SetPosition(const Value: TPoint);
		function GetNodeSize: TSize;
		procedure SetFixedSize(const Value: TSize);
	public
		constructor Create(AFixedSize: TSize); reintroduce;
		destructor Destroy; override;

		property NodeSize: TSize read GetNodeSize;
		property Parent: TLTObject read FParent write SetParent;
		property Position: TPoint read FPosition write SetPosition;
		property FixedSize: TSize read FFixedSize write SetFixedSize;
	end;

	TLTPaneNode = class(TLTNode)
	private
		FPane: TLTPane;
		procedure SetPane(const Value: TLTPane);
	public
		constructor Create(Pane: TLTPane; AFixedSize: TSize); reintroduce;
		destructor Destroy; override;

		property Pane: TLTPane read FPane write SetPane;

		procedure Realign; override;
		procedure Paint(Canvas: TCanvas); override;
		procedure Invalidate; override;
	end;

	TLTCntrNode = class;

	TLTLineItem = class
	private
		function GetCells(Idx: Integer): TLTNode;
	protected
		FItems: DArray;
		FOwner: TLTCntrNode;
		function GetOffset: Integer; virtual; abstract;
		procedure SetOffset(const Value: Integer); virtual; abstract;
	public
		constructor Create(AOwner: TLTCntrNode);
		destructor Destroy; override;

		function CalcOffset: Integer; virtual; abstract;

		property Cells[Idx: Integer]: TLTNode read GetCells; default;

		property Offset: Integer read GetOffset write SetOffset;
	end;

	TLTRow = class(TLTLineItem)
	private
		FRow: Integer;
		procedure SetRow(const Value: Integer);
		procedure SetHeight(const Value: Integer);
		function GetHeight: Integer;
		procedure SetFixedHeight(const Value: Integer);
		function GetFixedHeight: Integer;
	protected
		function GetOffset: Integer; override;
		procedure SetOffset(const Value: Integer); override;
	public
		constructor Create(AOwner: TLTCntrNode; Row: Integer);
		destructor Destroy; override;

		function CalcOffset: Integer; override;

		property Row: Integer read FRow write SetRow;
		property Height: Integer read GetHeight write SetHeight;
		property FixedHeight: Integer read GetFixedHeight write SetFixedHeight;
	end;

	TLTCol = class(TLTLineItem)
	private
		FCol: Integer;
		procedure SetCol(const Value: Integer);
		procedure SetWidth(const Value: Integer);
		function GetWidth: Integer;
		function GetFixedWidth: Integer;
		procedure SetFixedWidth(const Value: Integer);
	protected
		function GetOffset: Integer; override;
		procedure SetOffset(const Value: Integer); override;
	public
		constructor Create(AOwner: TLTCntrNode; Col: Integer);
		destructor Destroy; override;

		function CalcOffset: Integer; override;

		property Col: Integer read FCol write SetCol;
		property Width: Integer read GetWidth write SetWidth;
		property FixedWidth: Integer read GetFixedWidth write SetFixedWidth;
	end;

	TLTCntrNode = class(TLTNode)
	private
		FColCount: Integer;
		FRowCount: Integer;

		FNodes: DMap;
		function GetNodes(Col, Row: Integer): TLTNode;
		procedure SetNodes(Col, Row: Integer; const Value: TLTNode);
		function GetCols(Col: Integer): TLTCol;
		function GetRows(Row: Integer): TLTRow;
		procedure RealignRows;
		procedure RealignCols;
		procedure PutNode(SourceNodeCoords:TPoint; SourceNode: TLTNode);
	public
		property RowCount: Integer read FRowCount;
		property ColCount: Integer read FColCount;
		property PlainNodes: DMap read FNodes;

		constructor Create(AColCount, ARowCount: Integer; AFixedSize: TSize); reintroduce; overload;
		constructor Create(AColCount, ARowCount: Integer); reintroduce; overload;

		destructor Destroy; override;

		procedure Clean;

		procedure Realign; override;
		procedure MakeSameWidth;
		procedure MakeSameHeight;
		procedure MakeTheSame;

		function CreatePaneNode(Col, Row: Integer; Pane: TLTPane; AFixedSize: TSize): TLTPaneNode; overload;
		function CreatePaneNode(Col, Row: Integer; Pane: TLTPane): TLTPaneNode; overload;

		function CreateCntrNode(Col, Row: Integer; AColCount, ARowCount: Integer; AFixedSize: TSize): TLTCntrNode; overload;
		function CreateCntrNode(Col, Row: Integer; AColCount, ARowCount: Integer): TLTCntrNode; overload;

		property Nodes[Col, Row: Integer]: TLTNode read GetNodes write SetNodes; default;
		property Rows[Row: Integer]: TLTRow read GetRows;
		property Cols[Col: Integer]: TLTCol read GetCols;

		procedure Paint(Canvas: TCanvas); override;
		procedure Invalidate; override;
	end;

	TLayout = class(TLTObject)
	private
		FBoundsRect: TRect;
		FRoot: TLTNode;
		FOwner: TCustomControl;
		procedure SetRoot(const Value: TLTNode);
	protected
		procedure SetBoundsRect(const Value: TRect); override;
	public
		constructor Create(Owner: TCustomControl);
		destructor Destroy; override;

		property Root: TLTNode read FRoot write SetRoot;

		procedure Realign; override;
		procedure Paint(Canvas: TCanvas); override;
		procedure Invalidate; override;
	end;

	TTLTPane_class = class of TLTPane;
	TLTNode_class = class of TLTNode;

implementation

uses SysUtils, Math;

{ TRTable }

constructor TRTable.Create(rcRect: TRect; RowCount, ColCount: Integer);
begin
  FRowCount := RowCount;
  FColCount := ColCount;
  FBoundsRect := rcRect;
end;

destructor TRTable.Destroy;
begin

  inherited;
end;

function TRTable.GetCell(Col, Row: Integer): TRect;
var Height, Width: Integer;
begin
  Height := RHeight(FBoundsRect);
  Width := RWidth(FBoundsRect);

  Result.Top := Row * Height div FRowCount + FBoundsRect.Top;
  Result.Bottom := (Row + 1) * Height div FRowCount + FBoundsRect.Top;
  Result.Left := Col * Width div FColCount + FBoundsRect.Left;
  Result.Right := (Col + 1) * Width div FColCount + FBoundsRect.Left;
end;

procedure TRTable.SetBoundsRect(const Value: TRect);
begin
  FBoundsRect := Value;
end;

procedure TRTable.SetCell(Col, Row: Integer; const Value: TRect);
begin
  // Do nothing. Do not call this proc directly.
end;

{ TLTNode }

constructor TLTNode.Create(AFixedSize: TSize);
begin
  FBoundsRect := Rect(0,0,0,0);
  FFixedSize := AFixedSize;
end;

destructor TLTNode.Destroy;
begin

	inherited;
end;

procedure TLTNode.SetParent(const Value: TLTObject);
begin
  FParent := Value;
end;

procedure TLTNode.SetPosition(const Value: TPoint);
begin
  {TODO: Should be much more complex. But this proc is not used so far.}
  FPosition := Value;
end;

function TLTNode.GetNodeSize: TSize;
begin
  Result := RectSize(BoundsRect);
end;

procedure TLTNode.SetFixedSize(const Value: TSize);
begin
  FFixedSize := Value;
end;

{ TLayout }

constructor TLayout.Create(Owner: TCustomControl);
begin
	FOwner := Owner;
  FBoundsRect := Rect(0,0,0,0);

  FRoot := nil;
end;

destructor TLayout.Destroy;
begin
  if Assigned(FRoot) then
    FreeAndNil(FRoot);

  inherited;
end;

procedure TLayout.Invalidate;
begin
  inherited;

	FOwner.Invalidate;
end;

procedure TLayout.Paint(Canvas: TCanvas);
begin
  //BoundsRect := FOwner.ClientRect;
  FRoot.Paint(Canvas);
end;

procedure TLayout.Realign;
begin
  FRoot.Realign;
end;

procedure TLayout.SetBoundsRect(const Value: TRect);
begin
  inherited;
  FRoot.BoundsRect := BoundsRect;
end;

procedure TLayout.SetRoot(const Value: TLTNode);
begin
  FRoot := Value;
end;

{ TLTPane }

constructor TLTPane.Create(AOwner: TComponent; ANode: TLTPaneNode);
begin
	inherited Create(AOwner);
	FNode := ANode;
end;

destructor TLTPane.Destroy;
begin

  inherited;
end;

function TLTPane.GetPaintArea: TRect;
begin
	Result := Rect(0,0,0,0);
	if Assigned(Node) then
		Result := Node.BoundsRect;
end;

procedure TLTPane.Invalidate;
begin
	if Assigned(Node) then
		Node.Parent.Invalidate;
end;

procedure TLTPane.SetNode(const Value: TLTPaneNode);
begin
	FNode := Value;
	FNode.Pane := Self;
end;

{ TLTCntrNode }

constructor TLTCntrNode.Create(AColCount, ARowCount: Integer; AFixedSize: TSize);
begin
  inherited Create(AFixedSize);

  FRowCount := ARowCount;
	FColCount := AColCount;

  FNodes := DMap.Create;
end;

destructor TLTCntrNode.Destroy;
begin
	Clean;

	FreeAndNil(FNodes);

	inherited;
end;

function TLTCntrNode.GetCols(Col: Integer): TLTCol;
begin
  Result := TLTCol.Create(Self, Col);
end;

function TLTCntrNode.GetNodes(Col, Row: Integer): TLTNode;
var i: DIterator;
begin
	Result := nil;

  i := FNodes.locate([Pt2Key(Point(Col, Row))]);

  if not atEnd(i) then
    Result := getObject(i) as TLTNode;
end;

function TLTCntrNode.GetRows(Row: Integer): TLTRow;
begin
  Result := TLTRow.Create(Self, Row);
end;

procedure TLTCntrNode.MakeSameHeight;
var r: integer;
    SingleCellHeight: integer;
    Row: TLTRow;
begin
  SingleCellHeight := NodeSize.cy div RowCount;

  for r := 0 to RowCount - 1 do
  begin
    Row := Rows[r];

    Row.Offset := r * SingleCellHeight;
    Row.Height := SingleCellHeight;

    Row.Free;
  end;
end;

procedure TLTCntrNode.MakeSameWidth;
var c: integer;
    SingleCellWidth: integer;
    Col: TLTCol;
begin
  SingleCellWidth := NodeSize.cx div ColCount;

  for c := 0 to ColCount - 1 do
  begin
    Col := Cols[c];

    Col.Offset := c * SingleCellWidth;
    Col.Width := SingleCellWidth;

    Col.Free;
  end;
end;

procedure TLTCntrNode.MakeTheSame;
begin
  MakeSameWidth;
  MakeSameHeight;
end;

procedure TLTCntrNode.Paint(Canvas: TCanvas);
var r,c: integer;
    Node: TLTNode;
begin
  for r := 0 to RowCount - 1 do
    for c := 0 to ColCount - 1 do
    begin
      Node := Nodes[c,r];
      if Assigned(Node) then
        Node.Paint(Canvas);
    end;
end;

procedure TLTCntrNode.RealignRows;
var r: integer;
    Row: TLTRow;
    IsFixedSizeRow: boolean;
    FixedHeight: Integer;

    StartingFixedRowCount: Integer;
    FinishingFixedRowCount: Integer;
    MiddleFixedRowCount: Integer;

    HeightAccumulator: Integer;

    EvenRowHeight: Integer;
    EvenRowCount: Integer;
begin
	StartingFixedRowCount := 0;
  FinishingFixedRowCount := 0;
  MiddleFixedRowCount := 0;

  IsFixedSizeRow := true;
	HeightAccumulator := 0;

	r := 0;

  // Process fixed rows located at start.
  while IsFixedSizeRow and (r <= RowCount - 1) do
	begin
		Row := Rows[r];

    FixedHeight := Row.FixedHeight;

		if FixedHeight <> 0 then
		begin
      Row.Offset := Row.Offset; // Do not remove this line!
      Row.Height := FixedHeight;
			//This code explicitly set offset based on previous row heights.

      Inc(HeightAccumulator, FixedHeight);
			Inc(StartingFixedRowCount);
    end
    else
      IsFixedSizeRow := false;

		Row.Free;

    Inc(r);
  end;

  IsFixedSizeRow := true;

  r := RowCount - 1;
  // Process fixed rows located at finish excluding fixed rows located at start.
  while IsFixedSizeRow and (r > StartingFixedRowCount - 1) do
  begin
    Row := Rows[r];

    FixedHeight := Row.FixedHeight;

    if FixedHeight <> 0 then
    begin
      //Row.Offset := Row.Offset; // Do not remove this line!
      Row.Height := FixedHeight;
      //This code explicitly set the offset based on previous row heights.

      Inc(HeightAccumulator, FixedHeight);
      Inc(FinishingFixedRowCount);
    end
    else
      IsFixedSizeRow := false;

    Row.Free;
    Dec(r);
  end;

	r := StartingFixedRowCount;
	// Processing all fixed rows located between fixed starting and finishing strips.
	while (r <= RowCount - FinishingFixedRowCount - 1) do
	begin
		Row := Rows[r];

    FixedHeight := Row.FixedHeight;
    If FixedHeight > 0 then
    begin
      // This row is fixed one. Increase fixed height accumulator.
      Inc(HeightAccumulator, FixedHeight);
      Inc(MiddleFixedRowCount);
    end;

    Row.Free;
    Inc(r);
  end;

  // Calculate single even row height.
	EvenRowCount := RowCount - (StartingFixedRowCount + FinishingFixedRowCount + MiddleFixedRowCount);
	if EvenRowCount = 0 then EvenRowCount := 1;
	EvenRowHeight := (RHeight(BoundsRect) - HeightAccumulator) div EvenRowCount;

  // Arrange the rows depending on gathered data.
  for r := 0 to RowCount - 1 do
  begin
    Row := Rows[r];

		Row.Offset := Row.CalcOffset; // Set offset based on previous rows.

    FixedHeight := Row.FixedHeight;

    if FixedHeight > 0 then
    begin
			Row.Height := FixedHeight;  // For each fixed row we should set height to FixedHeight.
		end
    else
			Row.Height := EvenRowHeight; // ... else make their heights equal-sized.

    Row.Free;
  end;
end;

procedure TLTCntrNode.RealignCols;
var c: integer;
    Col: TLTCol;
    IsFixedSizeCol: boolean;
    FixedWidth: Integer;

    StartingFixedColCount: Integer;
    FinishingFixedColCount: Integer;
    MiddleFixedColCount: Integer;

    WidthAccumulator: Integer;

    EvenColWidth: Integer;
    EvenColCount: Integer;
begin
	StartingFixedColCount := 0;
	FinishingFixedColCount := 0;
	MiddleFixedColCount := 0;

	IsFixedSizeCol := true;
	WidthAccumulator := 0;

	c := 0;

	// Process fixed Cols located at start.
	while IsFixedSizeCol and (c <= ColCount - 1) do
	begin
		Col := Cols[c];

		FixedWidth := Col.FixedWidth;

		if FixedWidth <> 0 then
		begin
			Col.Offset := Col.Offset; // Do not remove this line!
			Col.Width := FixedWidth;
			//This code explicitly set offset based on previous Col Widths.

			Inc(WidthAccumulator, FixedWidth);
			Inc(StartingFixedColCount);
		end
		else
			IsFixedSizeCol := false;

		Col.Free;

		Inc(c);
	end;

	IsFixedSizeCol := true;

	c := ColCount - 1;
	// Process fixed Cols located at finish excluding fixed Cols located at start.
	while IsFixedSizeCol and (c >= StartingFixedColCount) do
	begin
		Col := Cols[c];

		FixedWidth := Col.FixedWidth;

		if FixedWidth <> 0 then
		begin
			//Col.Offset := Col.Offset; // Do not remove this line!
			Col.Width := FixedWidth;
			//This code explicitly set the offset based on previous Col Widths.

			Inc(WidthAccumulator, FixedWidth);
			Inc(FinishingFixedColCount);
		end
		else
			IsFixedSizeCol := false;

		Col.Free;
		Dec(c);
	end;

	c := StartingFixedColCount;
	// Processing all fixed Cols located between fixed starting and finishing strips.
	while (c <= ColCount - FinishingFixedColCount - 1) do
	begin
		Col := Cols[c];

		FixedWidth := Col.FixedWidth;
		If FixedWidth > 0 then
		begin
			// This Col is fixed one. Increase fixed Width accumulator.
			Inc(WidthAccumulator, FixedWidth);
			Inc(MiddleFixedColCount);
		end;

		Col.Free;
		Inc(c);
	end;

	// Calculate single even Col Width.
	EvenColCount := ColCount - (StartingFixedColCount + FinishingFixedColCount + MiddleFixedColCount);
  if EvenColCount = 0 then EvenColCount := 1;
	EvenColWidth := (RWidth(BoundsRect) - WidthAccumulator) div EvenColCount;

	// Arrange the Cols depending on gathered data.
	for c := 0 to ColCount - 1 do
	begin
		Col := Cols[c];

    Col.Offset := Col.CalcOffset; // Set offset based on previous Cols.

    FixedWidth := Col.FixedWidth;

    if FixedWidth > 0 then
    begin
      Col.Width := FixedWidth;  // For each fixed Col set it Width to FixedWidth.
    end
    else
      Col.Width := EvenColWidth; // ... else distribute their Widths evenly.

    Col.Free;
  end;
end;

procedure TLTCntrNode.Realign;
begin
  RealignRows;
  RealignCols;
end;

procedure TLTCntrNode.SetNodes(Col, Row: Integer; const Value: TLTNode);
var i: DIterator;
		Node: TLTNode;
		Obj: TObject;
begin
	Value.Parent := Self;
	Value.Position := Point(Col, Row);

	i := FNodes.locate([Pt2Key(Point(Col, Row))]);

	if not atEnd(i) then
	begin
		// There is another instance.
		Obj := getObject(i);
		Node := Obj as TLTNode;
		FreeAndNil(Node); // Remove this object.
		put(i, [Value]); // replace old value with the new one.
	end
	else
		FNodes.putPair([Pt2Key(Point(Col, Row)), Value]); // All right.

  Realign;
end;

function TLTCntrNode.CreatePaneNode(Col, Row: Integer; Pane: TLTPane;
	AFixedSize: TSize): TLTPaneNode;
var Node: TLTPaneNode;
begin
	Node := TLTPaneNode.Create(Pane, AFixedSize);

	if Assigned(Pane) then
		Pane.Node := Node;  // cause otherwise we have nothing to set.

	Nodes[Col, Row] := Node;
	Result := Node;
end;

function TLTCntrNode.CreateCntrNode(Col, Row, AColCount, ARowCount: Integer;
  AFixedSize: TSize): TLTCntrNode;
var Node: TLTCntrNode;
begin
	Node := TLTCntrNode.Create(AColCount, ARowCount, AFixedSize);
  Nodes[Col, Row] := Node;
  Result := Node;
end;

function TLTCntrNode.CreateCntrNode(Col, Row, AColCount,
  ARowCount: Integer): TLTCntrNode;
begin
  Result := CreateCntrNode(Col, Row, AColCount, ARowCount, Size(0,0));
end;

function TLTCntrNode.CreatePaneNode(Col, Row: Integer;
  Pane: TLTPane): TLTPaneNode;
begin
	Result := CreatePaneNode(Col, Row, Pane, Size(0,0));
end;

constructor TLTCntrNode.Create(AColCount, ARowCount: Integer);
begin
  Create(AColCount, ARowCount, Size(0,0));
end;

procedure TLTCntrNode.Clean;
var i: DIterator;
		Node: TLTNode;
begin
	if not Assigned(FNodes) then Exit;

	i := FNodes.start;

	while IterateOver(i) do
	begin
		Node := getObject(i) as TLTNode;
		Node.Free;
	end;

	FNodes.clear;
end;

procedure TLTCntrNode.PutNode(SourceNodeCoords: TPoint;
	SourceNode: TLTNode);
var CntrNode: TLTCntrNode;
begin
	//NewNode := TLTNode(SourceNode.ClassType.Create);
	//NewNode.Assign(SourceNode);
	if SourceNode is TLTPaneNode then
	begin
		// Create nodes with empty panes. They will be set later when panes
		CreatePaneNode(SourceNodeCoords.X, SourceNodeCoords.Y, nil, SourceNode.FixedSize)
	end
	else if SourceNode is TLTCntrNode then
	begin
		CntrNode := SourceNode as TLTCntrNode;
		CreateCntrNode(SourceNodeCoords.X, SourceNodeCoords.Y, CntrNode.ColCount, CntrNode.RowCount, SourceNode.FixedSize);
	end;
end;

procedure TLTCntrNode.Invalidate;
begin
  inherited;

	if Assigned(Parent) then
		Parent.Invalidate;
end;

{ TLTPaneNode }

constructor TLTPaneNode.Create(Pane: TLTPane; AFixedSize: TSize);
begin
  inherited Create(AFixedSize);

  FPane := Pane;
end;

destructor TLTPaneNode.Destroy;
begin
  //if Assigned(FPane) then
  //  FreeAndNil(FPane);

  inherited;
end;

procedure TLTPaneNode.Invalidate;
begin
	//inherited;

	Parent.Invalidate; //???
end;

procedure TLTPaneNode.Paint(Canvas: TCanvas);
begin
	if Assigned(Pane) then
		Pane.Paint(Canvas);
end;

procedure TLTPaneNode.Realign;
begin
  // We have to do nothing. Pane nodes will be realigned by their owners.
end;

procedure TLTPaneNode.SetPane(const Value: TLTPane);
begin
  FPane := Value;
end;

{ TLTObject }

procedure TLTObject.SetBoundsRect(const Value: TRect);
begin
	FBoundsRect := Value;
end;

{ TLTLineItem }

constructor TLTLineItem.Create(AOwner: TLTCntrNode);
begin
  FOwner := AOwner;
	FItems := nil;
end;

destructor TLTLineItem.Destroy;
begin
  if Assigned(FItems) then
    FreeAndNil(FItems);
  inherited;
end;

function TLTLineItem.GetCells(Idx: Integer): TLTNode;
begin
  Result := FItems.atAsObject(Idx) as TLTNode;
end;

{ TLTRow }

function TLTRow.CalcOffset: Integer;
var r: integer;
    ARow: TLTRow;
begin
  Result := 0;

  for r := 0 to FRow - 1 do
  begin
    ARow := FOwner.Rows[r];

    Inc(Result, ARow.Height);

    ARow.Free;
  end;

  Result := Result + FOwner.BoundsRect.Top;
end;

constructor TLTRow.Create(AOwner: TLTCntrNode; Row: Integer);
var c: integer;
begin
  inherited Create(AOwner);
  FRow := Row;

  FItems := DArray.createSize(AOwner.ColCount);

  for c := 0 to AOwner.ColCount - 1 do
  begin
     FItems.add([AOwner.Nodes[c,Row]]);
  end;
end;

destructor TLTRow.Destroy;
begin

  inherited;
end;

function TLTRow.GetFixedHeight: Integer;
var c: integer;
    FixedHeight, FixedHeightToBeChecked: Integer;
    Node: TLTNode;
begin
  Node := FItems.atAsObject(0) as TLTNode;

  if not Assigned(Node) then
  begin
    Result := 0;
    Exit;
  end;

  FixedHeight := Node.FixedSize.cy;

  for c := 1 to FItems.size - 1 do
  begin
    Node := FItems.atAsObject(c) as TLTNode;
    if not Assigned(Node) then continue;

    FixedHeightToBeChecked := Node.FixedSize.cy;

    if FixedHeight <> FixedHeightToBeChecked then
    begin
      Result := 0;
      Exit;
    end;
  end;

  Result := FixedHeight;
end;

function TLTRow.GetHeight: Integer;
var c: integer;
    MaxHeight: Integer;
    Height: Integer;
    Node: TLTNode;
begin
  MaxHeight := 0;
  for c := 0 to FItems.size - 1 do
  begin
    Node := FItems.atAsObject(c) as TLTNode;
    if not Assigned(Node) then continue;

    Height := RHeight(Node.BoundsRect);

    if Height > MaxHeight then
    begin
      MaxHeight := Height;
    end;
  end;

  Result := MaxHeight;
end;

function TLTRow.GetOffset: Integer;
var c: integer;
    Node: TLTNode;
begin
  Result := MaxInt;
  for c := 0 to FItems.size - 1 do
  begin
    Node := FItems.atAsObject(c) as TLTNode;
    if not Assigned(Node) then continue;

    Result := Min(Node.BoundsRect.Top, Result);
  end;
end;

procedure TLTRow.SetFixedHeight(const Value: Integer);
var c: integer;
    sz: TSize;
begin
  for c := 0 to FItems.size - 1 do
  begin
    sz := (FItems.atAsObject(c) as TLTNode).FixedSize;
    sz.cy := Value;
    (FItems.atAsObject(c) as TLTNode).FixedSize := sz;
  end;
end;

procedure TLTRow.SetHeight(const Value: Integer);
var c: integer;
    rcRect: TRect;
    Node: TLTNode;
begin
	for c := 0 to FItems.size - 1 do
  begin
    Node := FItems.atAsObject(c) as TLTNode;
    if Assigned(Node) then
    begin
      rcRect := Node.BoundsRect;
      SetRHeight(rcRect, Value);
			Node.BoundsRect := rcRect;

			Node.Realign;
			// Needs realign only if sizes were changed.
    end;
  end;
end;

procedure TLTRow.SetOffset(const Value: Integer);
var c: integer;
    Node: TLTNode;
    rcRect: TRect;
begin
  for c := 0 to FItems.size - 1 do
  begin
    Node := FItems.atAsObject(c) as TLTNode;

    if not Assigned(Node) then continue;

    rcRect := Node.BoundsRect;
    RPressToTop(rcRect, Value);
    Node.BoundsRect := rcRect;
  end;
end;

procedure TLTRow.SetRow(const Value: Integer);
var c: integer;
begin
  FRow := Value;

  for c := 0 to FItems.size - 1 do
  begin
    FItems.putAt(c, [FOwner.Nodes[c, Value]]);
  end;
end;

{ TLTCol }

function TLTCol.CalcOffset: Integer;
var c: integer;
    ACol: TLTCol;
begin
  Result := 0;

  for c := 0 to FCol - 1 do
  begin
    ACol := FOwner.Cols[c];

    Inc(Result, ACol.Width);

    ACol.Free;
  end;

  Result := Result + FOwner.BoundsRect.Left;
end;

constructor TLTCol.Create(AOwner: TLTCntrNode; Col: Integer);
var r: integer;
begin
  inherited Create(AOwner);
  FCol := Col;

  FItems := DArray.createSize(AOwner.RowCount);

  for r := 0 to AOwner.RowCount - 1 do
  begin
    FItems.putAt(r, [AOwner.Nodes[Col,r]]);
  end;
end;

destructor TLTCol.Destroy;
begin

  inherited;
end;

function TLTCol.GetFixedWidth: Integer;
var r: integer;
    FixedWidth, FixedWidthToBeChecked: Integer;
    Node: TLTNode;
begin
  Node := FItems.atAsObject(0) as TLTNode;

  if not Assigned(Node) then
  begin
    Result := 0;
    Exit;
  end;

  FixedWidth := Node.FixedSize.cx;

  for r := 1 to FItems.size - 1 do
  begin
    Node := FItems.atAsObject(r) as TLTNode;
    if not Assigned(Node) then continue;

    FixedWidthToBeChecked := Node.FixedSize.cx;

    if FixedWidth <> FixedWidthToBeChecked then
    begin
      Result := 0;
      Exit;
    end;
  end;

  Result := FixedWidth;
end;

function TLTCol.GetOffset: Integer;
var r: integer;
    Node: TLTNode;
begin
  Result := MaxInt;
  for r := 0 to FItems.size - 1 do
  begin
    Node := FItems.atAsObject(r) as TLTNode;

    if not Assigned(Node) then continue;

    Result := Min(Node.BoundsRect.Left, Result);
  end;
end;

function TLTCol.GetWidth: Integer;
var r: integer;
    MaxWidth: Integer;
    Width: Integer;
    Node: TLTNode;
begin
  MaxWidth := 0;
  for r := 0 to FItems.size - 1 do
  begin
    Node := FItems.atAsObject(r) as TLTNode;

    if not Assigned(Node) then continue;
    Width := RWidth(Node.BoundsRect);

    if Width > MaxWidth then
    begin
      MaxWidth := Width;
    end;
  end;

  Result := MaxWidth;
end;

procedure TLTCol.SetCol(const Value: Integer);
var r: integer;
begin
  FCol := Value;

  for r := 0 to FItems.size - 1 do
  begin
    FItems.putAt(r, [FOwner.Nodes[Value, r]]);
  end;
end;

procedure TLTCol.SetFixedWidth(const Value: Integer);
var r: integer;
    sz: TSize;
begin
  for r := 0 to FItems.size - 1 do
  begin
    sz := (FItems.atAsObject(r) as TLTNode).FixedSize;
    sz.cx := Value;
    (FItems.atAsObject(r) as TLTNode).FixedSize := sz;
  end;
end;

procedure TLTCol.SetOffset(const Value: Integer);
var r: integer;
    Node: TLTNode;
    rcRect: TRect;
begin
  for r := 0 to FItems.size - 1 do
  begin
    Node := FItems.atAsObject(r) as TLTNode;

    if not Assigned(Node) then continue;

    rcRect := Node.BoundsRect;
    RPressToLeft(rcRect, Value);
    Node.BoundsRect := rcRect;
  end;
end;

procedure TLTCol.SetWidth(const Value: Integer);
var r: integer;
    rcRect: TRect;
    Node: TLTNode;
begin
	for r := 0 to FItems.size - 1 do
  begin
    Node := FItems.atAsObject(r) as TLTNode;
    if Assigned(Node) then
    begin
      rcRect := Node.BoundsRect;
      SetRWidth(rcRect, Value);
			Node.BoundsRect := rcRect;
			Node.Realign;
			// Needs realign only if sizes were changed.
    end;
  end;
end;

{ TLayoutPane }

constructor TLayoutPane.Create(AOwner: TComponent; ANode: TLTPaneNode);
begin
	inherited Create(AOwner, ANode);
	SetSubComponent(true);
end;

end.
